# -*- coding:utf-8 -*-
def _getTarget():
    from alchemyjsonschema.dictify import ModelLookup
    return ModelLookup


def _makeOne(*args, **kwargs):
    return _getTarget()(*args, **kwargs)


def test_root_model():
    import alchemyjsonschema.tests.models as models
    from alchemyjsonschema.tests.models import Group
    target = _makeOne(models)
    result = target("Group")
    assert result == Group


def test_child__of__root_model():
    import alchemyjsonschema.tests.models as models
    from alchemyjsonschema.tests.models import User
    """
    Group.users -> [User]
    """
    target = _makeOne(models)
    target("Group")
    result = target("users")
    assert result == User


def test_child__of__child__of__root_model():
    import alchemyjsonschema.tests.models as models
    from alchemyjsonschema.tests.models import Group
    """
    Group.users -> [User]
    User.group -> Group
    """
    target = _makeOne(models)
    target("Group")
    target("users")
    result = target("group")
    assert result == Group



# -*- coding:utf-8 -*-
def _getTarget():
    from alchemyjsonschema import SchemaFactory
    return SchemaFactory


def _makeOne(*args, **kwargs):
    from alchemyjsonschema import (
        SingleModelWalker,
        DefaultClassfier
    )
    return _getTarget()(SingleModelWalker, DefaultClassfier)


# definition
import sqlalchemy as sa
import sqlalchemy.orm as orm
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class Group(Base):
    """model for test"""
    __tablename__ = "Group"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key")
    name = sa.Column(sa.String(255), default="", nullable=False)
    color = sa.Column(sa.Enum("red", "green", "yellow", "blue"))


class User(Base):
    __tablename__ = "User"

    pk = sa.Column(sa.Integer, primary_key=True, doc="primary key")
    name = sa.Column(sa.String(255), default="", nullable=True)
    group_id = sa.Column(sa.Integer, sa.ForeignKey(Group.pk), nullable=False)
    group = orm.relationship(Group, uselist=False, backref="users")


def test_type__is_object():
    target = _makeOne()
    result = target(Group)

    assert "type" in result
    assert result["type"] == "object"


def test_properties__are__all_of_columns():
    target = _makeOne()
    result = target(Group)

    assert "properties" in result
    assert list(sorted(result["properties"].keys())) == ["color", "name", "pk"]


def test_required__are__nullable_is_false_columns():
    target = _makeOne()
    result = target(Group)

    assert "required" in result
    assert list(sorted(result["required"])) == ["name", "pk"]


def test_title__id__model_class_name():
    target = _makeOne()
    result = target(Group)

    assert "title" in result
    assert result["title"] == Group.__name__


def test_description__is__docstring_of_model():
    target = _makeOne()
    result = target(Group)

    assert "description" in result
    assert result["description"] == Group.__doc__


def test_properties__all__this_is_slackoff_little_bit__all_is_all():   # hmm.
    target = _makeOne()
    result = target(Group)

    assert result["properties"] == {'color': {'maxLength': 6,
                                              'enum': ['red', 'green', 'yellow', 'blue'],
                                              'type': 'string'},
                                    'name': {'maxLength': 255, 'type': 'string'},
                                    'pk': {'description': 'primary key', 'type': 'integer'}}

# adaptive


def test__filtering_by__includes():
    target = _makeOne()
    result = target(Group, includes=["pk"])
    assert list(sorted(result["properties"].keys())) == ["pk"]


def test__filtering_by__excludes():
    target = _makeOne()
    result = target(Group, excludes=["pk"])
    assert list(sorted(result["properties"].keys())) == ["color", "name"]


def test__filtering_by__excludes_and_includes__conflict():
    import pytest
    from alchemyjsonschema import InvalidStatus

    target = _makeOne()
    with pytest.raises(InvalidStatus):
        target(Group, excludes=["pk"], includes=["pk"])

# overrides


def test__overrides__add():
    target = _makeOne()
    overrides = {"name": {"maxLength": 100}}
    result = target(Group, includes=["name"], overrides=overrides)
    result["properties"] == {"name": {"maxLength": 100, 'type': 'string'}}


def test__overrides__pop():
    from alchemyjsonschema import pop_marker
    target = _makeOne()
    overrides = {"name": {"maxLength": pop_marker}}
    result = target(Group, includes=["name"], overrides=overrides)
    result["properties"] == {"name": {'type': 'string'}}


def test__overrides__wrong_column():
    import pytest
    from alchemyjsonschema import InvalidStatus

    target = _makeOne()
    overrides = {"*missing-field*": {"maxLength": 100}}
    with pytest.raises(InvalidStatus):
        target(Group, includes=["name"], overrides=overrides)




# -*- coding:utf-8 -*-
import logging
logger = logging.getLogger(__name__)
from operator import getitem
from .compat import text_
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.relationships import RelationshipProperty
from functools import partial
import isodate
from .custom.format import (
    parse_time,  # more strict than isodate
    parse_date   # more strict
)
from jsonschema import validate
from . import InvalidStatus
import pytz


def datetime_rfc3339(ob):
    if ob.tzinfo:
        return ob.isoformat()
    return pytz.utc.localize(ob).isoformat()


def isoformat(ob):
    return ob.isoformat() + "Z"  # xxx


def isoformat0(ob):
    return ob.isoformat()


def raise_error(ob):
    raise Exception("convert failure. unknown format xxx of {}".format(ob))


def maybe_wrap(fn, default=None):
    def wrapper(ob):
        if ob is None:
            return default
        return fn(ob)
    return wrapper

# todo: look at required or not
jsonify_dict = {('string', None): maybe_wrap(text_),
                ('string', 'time',): maybe_wrap(isoformat),
                ('number', None): maybe_wrap(float),
                ('integer', None): maybe_wrap(int),
                ('boolean', None): maybe_wrap(bool),
                ('string', 'date-time'): maybe_wrap(datetime_rfc3339),
                ('string', 'date'): maybe_wrap(isoformat0),
                ('xxx', None): raise_error}


normalize_dict = {('string', None): maybe_wrap(text_),
                  ('string', 'time',): maybe_wrap(parse_time),
                  ('number', None): maybe_wrap(float),
                  ('integer', None): maybe_wrap(int),
                  ('boolean', None): maybe_wrap(bool),
                  ('string', 'date-time'): maybe_wrap(isodate.parse_datetime),
                  ('string', 'date'): maybe_wrap(parse_date),
                  ('xxx', None): raise_error}


def jsonify_of(ob, name, type_, registry=jsonify_dict):
    try:
        convert_fn = registry[type_]
        return convert_fn(getattr(ob, name))
    except KeyError:
        raise Exception("convert {} failure. unknown format {} of {}".format(name, type_, ob))


def normalize_of(ob, name, type_, registry=normalize_dict):
    try:
        convert_fn = registry[type_]
        return convert_fn(ob[name])
    except KeyError:
        raise Exception("convert {} failure. unknown format {} of {}".format(name, type_, ob))


def attribute_of(ob, name, type_):
    return getattr(ob, name)


def dictify(ob, schema, convert=attribute_of, getter=getattr):
    return dictify_properties(ob, schema["properties"], convert=convert, getter=getter)


def jsonify(ob, schema, convert=jsonify_of, getter=getattr, registry=jsonify_dict):
    convert = partial(convert, registry=registry)
    return dictify_properties(ob, schema["properties"], convert=convert, getter=getter)


def normalize(ob, schema, convert=normalize_of, getter=getitem, registry=normalize_dict):
    convert = partial(convert, registry=registry)
    return dictify_properties(ob, schema["properties"], convert=convert, getter=getter)


def dictify_properties(ob, properties, convert, getter):
    D = {}
    for k, v in properties.items():
        D[k] = _dictify(ob, k, v, convert, getter)
    return D


def _dictify(ob, name, schema, convert, getter):
    type_ = schema["type"]
    if type_ == "array":
        return [dictify_properties(e, schema["items"], convert, getter) for e in getter(ob, name)]
    elif type_ == "object":
        return dictify_properties(getter(ob, name), schema["properties"], convert, getter)
    else:
        return convert(ob, name, (type_, schema.get("format")))


class ModelLookup(object):
    def __init__(self, module):
        self.module = module
        self.name_stack = []
        self.inspect_stack = []

    def __call__(self, name):
        if not self.name_stack:
            self.name_stack.append(name)
            model = getattr(self.module, name)
            self.inspect_stack.append(inspect(model))
            return model
        else:
            self.name_stack.append(name)
            prop = self.inspect_stack[-1].get_property(name)
            assert isinstance(prop, RelationshipProperty)
            mapper = prop.mapper
            model = mapper.class_
            self.inspect_stack.append(mapper)
            return model

    def pop(self):
        name = self.name_stack.pop()
        return name, self.inspect_stack.pop()


class ComposedModule(object):
    def __init__(self, *modules):
        self.modules = set(modules)

    def __getattr__(self, k):
        for m in self.modules:
            if hasattr(m, k):
                return getattr(m, k)


# objectify
def objectify(params, schema, modellookup, strict=True):
    model_class = modellookup(schema["title"])
    result = model_class(**objectify_propperties(params, schema["properties"], modellookup))
    modellookup.pop()
    if strict:
        for k in schema["required"]:
            if getattr(result, k) is None:
                raise InvalidStatus("{}.{} is None. this is required.".format(model_class, k))
    assert modellookup.name_stack == []
    return result


def objectify_propperties(params, properties, modellookup):
    D = {}
    for k, schema in properties.items():
        D[k] = _objectify(params, k, schema, modellookup)
    return D


def _objectify(params, name, schema, modellookup):
    type_ = schema.get("type")
    if type_ == "array":
        return [_objectify_subobject(e, name, schema["items"], modellookup) for e in params[name]]
    elif type_ is None:  # object
        submodel = modellookup(name)
        result = submodel(**{k: _objectify(params[name], k, schema[k], modellookup) for k in schema})
        modellookup.pop()
        return result
    else:
        return params.get(name)


def _objectify_subobject(params, name, schema, modellookup):
    submodel = modellookup(name)
    result = submodel(**{k: _objectify(params, k, v, modellookup) for k, v in schema.items()})
    modellookup.pop()
    return result


class ErrorFound(Exception):  # xxx:
    pass


def validate_all(data, validator):
    errors = []
    for e in validator.iter_errors(data):
        errors.append(dict(name=e.path[0], reason=e.validator))
    if errors:
        raise ErrorFound(errors)
    return data



# -*- coding:utf-8 -*-
def _getTarget():
    from alchemyjsonschema import SchemaFactory
    return SchemaFactory


def _makeOne(*args, **kwargs):
    from alchemyjsonschema import (
        SingleModelWalker,
        DefaultClassfier
    )
    return _getTarget()(SingleModelWalker, DefaultClassfier)


# definition
from sqlalchemy.types import Integer, BigInteger
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class Identifier(Integer):
    def __init__(self, inner):
        self.inner = inner

    def _compiler_dispatch(self, visitor, **kw):
        return self.inner._compiler_dispatch(visitor, **kw)

    def is_not_support_big_integer(self):
        return True

    def get_dbapi_type(self, dbapi):
        return self.inner.get_dbapi_type(dbapi)

    @property
    def python_type(self):
        return self.inner.python_type

    @property
    def _expression_adaptations(self):
        return self.inner._expression_adaptations


class MyModel(Base):
    __tablename__ = "MyModel"
    pk = sa.Column(Identifier(BigInteger), primary_key=True)


class MyModel2(Base):
    __tablename__ = "MyModel2"
    pk = sa.Column(Identifier(Integer), primary_key=True)


def test_it():
    target = _makeOne()
    result = target(MyModel)
    assert result["properties"] == {"pk": {"type": "integer"}}


def test_it2():
    target = _makeOne()
    result = target(MyModel)
    assert result["properties"] == {"pk": {"type": "integer"}}



# -*- coding:utf-8 -*-
def _callFUT(*args, **kwargs):
    from alchemyjsonschema.dictify import objectify
    return objectify(*args, **kwargs)


def test_it__simple():
    from alchemyjsonschema import SchemaFactory, SingleModelWalker
    from alchemyjsonschema.dictify import ModelLookup
    import alchemyjsonschema.tests.models as models
    from datetime import datetime

    factory = SchemaFactory(SingleModelWalker)
    user_schema = factory(models.User)

    created_at = datetime(2000, 1, 1)
    user_dict = dict(pk=1, name="foo", created_at=created_at, group_id=10)
    modellookup = ModelLookup(models)

    result = _callFUT(user_dict, user_schema, modellookup)

    assert isinstance(result, models.User)
    assert result.pk == 1
    assert result.name == "foo"
    assert result.created_at == datetime(2000, 1, 1)
    assert result.group_id == 10


def test_it__strict_true__then__required_are_notfound__error_raised():
    from alchemyjsonschema import SchemaFactory, SingleModelWalker, InvalidStatus
    from alchemyjsonschema.dictify import ModelLookup
    import alchemyjsonschema.tests.models as models
    from datetime import datetime
    import pytest

    factory = SchemaFactory(SingleModelWalker)
    user_schema = factory(models.User)

    created_at = datetime(2000, 1, 1)
    user_dict = dict(name="foo", created_at=created_at)  # pk is not found
    modellookup = ModelLookup(models)

    with pytest.raises(InvalidStatus):
        _callFUT(user_dict, user_schema, modellookup, strict=True)


def test_it__strict_false__then__required_are_notfound__ok():
    from alchemyjsonschema import SchemaFactory, SingleModelWalker
    from alchemyjsonschema.dictify import ModelLookup
    import alchemyjsonschema.tests.models as models
    from datetime import datetime

    factory = SchemaFactory(SingleModelWalker)
    user_schema = factory(models.User)

    created_at = datetime(2000, 1, 1)
    user_dict = dict(name="foo", created_at=created_at)  # pk is not found
    modellookup = ModelLookup(models)

    result = _callFUT(user_dict, user_schema, modellookup, strict=False)

    assert isinstance(result, models.User)
    assert result.pk is None
    assert result.name == "foo"
    assert result.created_at == datetime(2000, 1, 1)
    assert result.group_id is None


def test_it_complex():
    from alchemyjsonschema import SchemaFactory, AlsoChildrenWalker
    from alchemyjsonschema.dictify import ModelLookup
    import alchemyjsonschema.tests.models as models
    from datetime import datetime

    factory = SchemaFactory(AlsoChildrenWalker)
    user_schema = factory(models.User)

    created_at = datetime(2000, 1, 1)
    created_at2 = datetime(2001, 1, 1)
    group_dict = dict(name="ravenclaw", color="blue", created_at=created_at2)
    user_dict = dict(name="foo", created_at=created_at, group=group_dict)  # pk is not found
    modellookup = ModelLookup(models)

    result = _callFUT(user_dict, user_schema, modellookup, strict=False)

    assert isinstance(result, models.User)
    assert result.pk is None
    assert result.name == "foo"
    assert result.created_at == datetime(2000, 1, 1)
    assert result.group_id is None

    assert isinstance(result.group, models.Group)
    assert result.group.name == "ravenclaw"
    assert result.group.color == "blue"
    assert result.group.created_at == created_at2

    assert modellookup.name_stack == []



# -*- coding:utf-8 -*-
def _callFUT(*args, **kwargs):
    from alchemyjsonschema.dictify import dictify
    return dictify(*args, **kwargs)


def test_it__dictify():
    from alchemyjsonschema import SchemaFactory, AlsoChildrenWalker
    from alchemyjsonschema.tests.models import Group, User
    from datetime import datetime

    factory = SchemaFactory(AlsoChildrenWalker)
    group_schema = factory(Group)

    created_at = datetime(2000, 1, 1)
    users = [User(name="foo", created_at=created_at),
             User(name="boo", created_at=created_at)]
    group = Group(name="ravenclaw", color="blue", users=users, created_at=created_at)

    result = _callFUT(group, group_schema)
    assert result == {'pk': None, 'color': 'blue', 'name': 'ravenclaw', "created_at": created_at,
                      'users': [{'pk': None, 'name': 'foo', "created_at": created_at},
                                {'pk': None, 'name': 'boo', "created_at": created_at}]}


def test_it__jsonify():
    from alchemyjsonschema import SchemaFactory, AlsoChildrenWalker
    from alchemyjsonschema.tests.models import Group, User
    from datetime import datetime
    from alchemyjsonschema.dictify import jsonify_of

    factory = SchemaFactory(AlsoChildrenWalker)
    group_schema = factory(Group)

    created_at = datetime(2000, 1, 1)
    users = [User(name="foo", created_at=created_at),
             User(name="boo", created_at=created_at)]
    group = Group(name="ravenclaw", color="blue", users=users, created_at=created_at)

    result = _callFUT(group, group_schema, convert=jsonify_of)
    assert result == {'name': 'ravenclaw', 'created_at': '2000-01-01T00:00:00+00:00', 'color': 'blue', 'pk': None,
                      'users': [{'name': 'foo', 'created_at': '2000-01-01T00:00:00+00:00', 'pk': None},
                                {'name': 'boo', 'created_at': '2000-01-01T00:00:00+00:00', 'pk': None}]}
